#include <stdio.h>
#include <string.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "as_tree.h"
#include "quad.h"
#include "peephole.h"

static void optim_label_reduct(struct quad **quads, unsigned int *size);

static struct {
	void (*optim_handler)(struct quad **, unsigned int *);
} optims[] = {
	{ optim_label_reduct },
	{ NULL }
};

static opcode_t invert_opcode(opcode_t original)
{
	struct {
		opcode_t original;
		opcode_t invert;
	} ops[] = {
		{ OPCODE_LT, OPCODE_GE },
		{ OPCODE_GT, OPCODE_LE },
		{ OPCODE_LE, OPCODE_GT },
		{ OPCODE_GE, OPCODE_LT },
		{ OPCODE_NE, OPCODE_EQ },
		{ OPCODE_EQ, OPCODE_NE },                                                                                                                        
		{ -1, -1 }
        };

	unsigned int i;

	for (i = 0; ops[i].original != -1; i++) {
		if (ops[i].original == original)
			return ops[i].invert;
	}

	return OPCODE_UNKNOWN;
}

static void optim_label_reduct(struct quad **quads, unsigned int *size)
{
	struct quad_goto *goto_quad;
	struct quad_label *label_quad;
	struct quad_if *if_quad;
	unsigned int *stat;
	unsigned int nstat, i, n;

	stat = NULL;
	nstat = 0;
	n = *size;
	/*
	 * Count references for each label. Later we will remove redundant
	 * jumps and decrement label statistic counters. The last step
	 * is to remove label quads that reference counters are 0.
	 */
	for (i = 0; i < n; i++) {
		int ln;
		switch (quads[i]->quad_type) {
		case QUAD_IF:
			if_quad = (struct quad_if *)quads[i];
			ln = if_quad->ln;
			break;
		case QUAD_GOTO:
			goto_quad = (struct quad_goto *)quads[i];
			ln = goto_quad->ln;
			break;
		default:
			continue;
		}

		if (ln >= nstat) {
			stat = xrealloc0(stat, sizeof(unsigned int)*nstat,
					sizeof(unsigned int)*(nstat+ln+1));
			nstat += ln+1;
		}
		++stat[ln];
	}
	/* Nothing was found, no reasons to waste time. */
	if (stat == NULL)
		return;
	/*
	 * Find the following pattern:
	 *
	 *	goto L3
	 *	:L3
	 * Free the goto quad, decrement the label statistic and shift
	 * rest array elements.
 	 */
	for (i = 0; i < n-1; i++) {
		if (quads[i]->quad_type != QUAD_GOTO ||
		    quads[i+1]->quad_type != QUAD_LABEL) 
			continue;

		goto_quad = (struct quad_goto *)quads[i];
		label_quad = (struct quad_label *)quads[i+1];
		
		if (goto_quad->ln != label_quad->ln)
			continue;

		--stat[goto_quad->ln];
		quad_unref(QUAD(goto_quad));

		memmove(&quads[i], &quads[i+1], sizeof(struct quad *)*(n-i-1));
		--n;
	}
	/*
	 * Find the following pattern:
	 *
	 *	if val1 rel_op val2 goto Ln
	 *	goto Lk
	 *	:Ln
	 * Change a rel_op operation to the oposite one, Ln to Lk, decrement
	 * the label statistic, free the goto quad and shift the rest
	 * elements.
	 */
	for (i = 0; i < n-2; i++) {
		if (quads[i]->quad_type != QUAD_IF ||
		    quads[i+1]->quad_type != QUAD_GOTO ||
		    quads[i+2]->quad_type != QUAD_LABEL)
			continue;

		if_quad = (struct quad_if *)quads[i];
		goto_quad = (struct quad_goto *)quads[i+1];
		label_quad = (struct quad_label *)quads[i+2];

		if (if_quad->ln != label_quad->ln)
			continue;
		
		--stat[if_quad->ln];
		if_quad->ln = goto_quad->ln;
		if_quad->opcode = invert_opcode(if_quad->opcode);

		quad_unref(QUAD(goto_quad));
		memmove(&quads[i+1], &quads[i+2], sizeof(struct quad *)*(n-i-2));
		--n;
	}
	/*
	 * Free labels that have 0 references.
	 */
	for (i = 0; i < n; i++) {
		if (quads[i]->quad_type != QUAD_LABEL)
			continue;
		
		label_quad = (struct quad_label *)quads[i];
		if (stat[label_quad->ln] != 0)
			continue;

		quad_unref(QUAD(label_quad));
		memmove(&quads[i], &quads[i+1], sizeof(struct quad *)*(n-i));
		--n;
	}

	xfree(stat);
	*size = n;
}

void peephole_optimization(struct quad **quads, unsigned int *size)
{
	unsigned int i;

	for (i = 0; optims[i].optim_handler != NULL; i++)
		optims[i].optim_handler(quads, size);
}

